/*
 * $QNXLicenseC: 
 * Copyright 2008, QNX Software Systems.  
 *  
 * Licensed under the Apache License, Version 2.0 (the "License"). You  
 * may not reproduce, modify or distribute this software except in  
 * compliance with the License. You may obtain a copy of the License  
 * at: http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" basis,  
 * WITHOUT WARRANTIES OF ANY KIND, either express or implied. 
 * 
 * This file may contain contributions from others, either as  
 * contributors under the License or as licensors under other terms.   
 * Please review this entire file for other proprietary rights or license  
 * notices, as well as the QNX Development Suite License Guide at  
 * http://licensing.qnx.com/license-guide/ for other information. 
 * $
 */

//int first = 1;

#include "proto.h"
//#include "..\bsp-freescale-mx6q-sabrelite-src\src\hardware\startup\lib\public\arm\mx6x.h"

void i2c_off(i2c_dev_t *dev)
{
//	out16(dev->regbase + I2C_CTRREG_OFF, in16(dev->regbase + I2C_CTRREG_OFF) & ~CTRREG_IEN);
	out16(dev->regbase + I2C_CTRREG_OFF, 0);
	out16(dev->regbase + AM335x_I2C_STAT, 0x07FFF);
	if (uiVerbosity >= LOGGING_DEBUG_NORMAL)
		slogf(_SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_DEBUG1, "%s:i2c_i2c_off: (CTRREG 0x%X STSREG 0x%X)",
				pszProgram, in16(dev->regbase + I2C_CTRREG_OFF), in16(dev->regbase + I2C_STSREG_OFF));
}

int	iMaster = 0;
int iReset = 0;
int	iSlave = 0;

void i2c_reset(i2c_dev_t *dev)
{
	int 	timeout = 2000;
//	int		iDoTwice =1;
	int		status = in16(dev->regbase + I2C_STSREG_OFF);
	int		ctrl = in16(dev->regbase + I2C_CTRREG_OFF);
	if (uiVerbosity >= LOGGING_DEBUG_NORMAL)
		slogf(_SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_DEBUG1, "%s:i2c_reset: entry (CTRREG 0x%X STSREG 0x%X)",
				pszProgram, ctrl, status);

	iMaster = 0;
	iSlave = 0;
	if ( ( ctrl & CTRREG_IEN ) && !( status & AM335x_I2C_RESET_ERROR ) && ( iReset++ < 100 ) )
	{
//		out16(dev->regbase + AM335x_I2C_CNT, 0);
		out16( dev->regbase + AM335x_I2C_STAT, status & ( AM335x_I2C_ARDY | AM335x_I2C_NACK  | AM335x_I2C_AL | AM335x_I2C_BF ) );
    	nanospin_ns( 1000 );
		int statusnew = in16(dev->regbase + I2C_STSREG_OFF);
		if ( status != statusnew )
		{
			printf( "Clear Control 0x%X Status 0x%X to 0x%X\n", ctrl, status, statusnew  );
			return;
		}
		if ( !( statusnew & AM335x_I2C_RESET_ERROR ) )
		{
			printf( "No Error Control 0x%X Status 0x%X so return\n", ctrl, statusnew  );
			return;
		}
	}
	iReset = 0;
	printf( "Doing reset Control 0x%X Status 0x%X\n", ctrl, status );
	// Disable interrupts
//DoTwice:
	out16(dev->regbase + AM335x_I2C_IE_CLR, 0x0FF); // clear interrupts
	// Disable DMA
	out16(dev->regbase + AM335x_I2C_BUF, 0);
	// Disable test mode
	out16(dev->regbase + AM335x_I2C_SYSTEST, 0);

	if ( in16( dev->regbase + I2C_CTRREG_OFF) & CTRREG_IEN )
		out16(dev->regbase + I2C_CTRREG_OFF, 0);
	out16(dev->regbase + AM335x_I2C_STAT, 0x07FFF );  //in16(dev->regbase + I2C_STSREG_OFF) & AM335x_I2C_RESET );
//	if ( first )
	out16(dev->regbase + AM335x_I2C_SYSC, AM335x_I2C_SRST );
	out16(dev->regbase + I2C_CTRREG_OFF, CTRREG_IEN );
    while ( !(in16(dev->regbase + AM335x_I2C_SYSS) & AM335x_I2C_SYSS_RDONE) && --timeout)
    {
    	nanospin_ns( 1000 );
    }
	if(!timeout)
	{
		if (uiVerbosity >= LOGGING_ERRORS)
			slogf(_SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_ERROR, "%s:i2c_reset: entry (CTRREG 0x%X STSREG 0x%X)",
					pszProgram, in16(dev->regbase + I2C_CTRREG_OFF), in16(dev->regbase + I2C_STSREG_OFF));
		printf( "Timeout I2c from reset\n" );
		exit(1);
	}
	out16(dev->regbase+AM335x_I2C_SYSC, AM335x_I2C_CLKACT_SYS | AM335x_I2C_IDLEMODE_SMART | AM335x_I2C_ENAWAKEUP | AM335x_I2C_AUTOIDLE ); //Smart idle mode, Fclock, Enable Wakeup, autoidle
	/*
	 * Enabling all wakup sources to stop I2C freezing on
	 * WFI instruction.
	 * REVISIT: Some wkup sources might not be needed.
	 */
	out16(dev->regbase + AM335x_I2C_WE, AM335x_I2C_WE_ALL );
	out16(dev->regbase + I2C_CTRREG_OFF, 0 );

	/* Set I2C bus speed */
//	if ( first )
	i2c_set_bus_speed( dev, dev->speed, NULL );

    /* Set Own Address */
	out16(dev->regbase + I2C_ADRREG_OFF, dev->own_addr );
	out16(dev->regbase + AM335x_I2C_SBLOCK, 0 );					// Tell it to only use OA0
//	printf( "Enable I2c from reset\n" );
/*
	if ( first )
	{
		first = 0;
		sleep( 30 );
	}
	if ( iDoTwice-- )
		goto DoTwice;
*/
	out16(dev->regbase + I2C_CTRREG_OFF, CTRREG_IEN );
	dev->restart = 0;
//	delay(10);
	sleep(3);
	if (uiVerbosity >= LOGGING_DEBUG_NORMAL)
		slogf(_SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_DEBUG1, "%s:i2c_reset: exit (CTRREG 0x%X STSREG 0x%X)",
				pszProgram, in16(dev->regbase + I2C_CTRREG_OFF), in16(dev->regbase + I2C_STSREG_OFF));
}

int i2c_wait_bus_not_busy(i2c_dev_t *dev)
{
    unsigned        tries = 1000000;
	if (uiVerbosity >= LOGGING_DEBUG_FINE)
		slogf(_SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_DEBUG2, "%s:i2c_wait_bus_not_busy: entry (CTRREG 0x%X STSREG 0x%X)",
				pszProgram, in16(dev->regbase + I2C_CTRREG_OFF), in16(dev->regbase + I2C_STSREG_OFF));
	if(dev->restart)
	{
		if (uiVerbosity >=  LOGGING_DEBUG_FINE)
			slogf( _SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_DEBUG2,
					"%s:i2c_wait_bus_not_busy: dev->restart != 0 so return 0", pszProgram);
		return 0;
	}
	while (in16(dev->regbase + I2C_STSREG_OFF) & STSREG_IBB)
	{
		if (tries-- == 0)
		{
			if (uiVerbosity >=  LOGGING_DEBUG_FINE)
				slogf( _SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_DEBUG2,
						"%s:i2c_wait_bus_not_busy: wait bus idle failed (CTRREG 0x%X STSREG 0x%X)",
						pszProgram, in16(dev->regbase + I2C_CTRREG_OFF), in16(dev->regbase + I2C_STSREG_OFF));
			/* reset the controller to see if it's able to recover*/
			i2c_reset(dev);
			//try again to see if it's OK now after reset
			if(in16(dev->regbase + I2C_STSREG_OFF) & STSREG_IBB)
			{
//				delay(1);
				if (uiVerbosity >=  LOGGING_DEBUG_FINE)
					slogf( _SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_DEBUG2,
							"%s:i2c_wait_bus_not_busy: STSREG_IBB still set, return -1 (CTRREG 0x%X STSREG 0x%X)",
							pszProgram, in16(dev->regbase + I2C_CTRREG_OFF), in16(dev->regbase + I2C_STSREG_OFF));
				return -1;
			}
			break;
		}
	}
	if (uiVerbosity >= LOGGING_DEBUG_FINE)
		slogf(_SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_DEBUG2, "%s:i2c_wait_bus_not_busy: return 0 (CTRREG 0x%X STSREG 0x%X)",
				pszProgram, in16(dev->regbase + I2C_CTRREG_OFF), in16(dev->regbase + I2C_STSREG_OFF));
	return 0;
}

#include <sys/time.h>
#include <time.h>

#ifdef sws
uint32_t i2c_master_wait_status(i2c_dev_t *dev)
{
	uint16_t	status1;

	iMaster = 1;
	if (uiVerbosity >= LOGGING_DEBUG_FINE)
		slogf(_SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_DEBUG2, "%s:i2c_master_wait_status: entry (CTRREG 0x%X STSREG 0x%X)",
				pszProgram, in16(dev->regbase + I2C_CTRREG_OFF), in16(dev->regbase + I2C_STSREG_OFF));
/*
    struct sigaction sa;
    int s, flags = 0;
    sigemptyset(&sa.sa_mask);
    sa.sa_flags = 0;
    sa.sa_handler = sigintHandler;
*/
    struct timespec start, stop, temp, request;
	clock_gettime( CLOCK_MONOTONIC, &start);
	do
	{
		//if ( (status = in16(dev->regbase + I2C_STSREG_OFF)) & STSREG_IIF )
		if ( (status1 = in16(dev->regbase + I2C_STSREG_OFF)) & AM335x_I2C_STAT_MASK )
		{
			// Clear the interrupt
//			out16(dev->regbase + I2C_STSREG_OFF, ( status1 & AM335x_I2C_INT_MASK ) );
//			out16(dev->regbase + I2C_STSREG_OFF, status = in16(dev->regbase + I2C_STSREG_OFF) & ~(STSREG_IIF));
//				InterruptUnmask(dev->intr, dev->master_iid);
			if (uiVerbosity >= LOGGING_DEBUG_FINE)
				slogf(_SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_DEBUG2, "%s:i2c_master_wait_status: STSREG_IIF set return STSREG 0x%X (CTRREG 0x%X STSREG 0x%X)",
						pszProgram, status1, in16(dev->regbase + I2C_CTRREG_OFF), in16(dev->regbase + I2C_STSREG_OFF));
//			timer_delete(timer_id);
			iMaster = 0;
			return status1;
		}
		request.tv_sec  = 0;
		request.tv_nsec = 1000;
//		clock_nanosleep(CLOCK_MONOTONIC, 0, &request, NULL);
		clock_gettime( CLOCK_MONOTONIC, &stop);
		if ((stop.tv_nsec-start.tv_nsec)<0)
		{
			temp.tv_sec = stop.tv_sec-start.tv_sec-1;
			temp.tv_nsec = 1000000000+stop.tv_nsec-start.tv_nsec;
		}
		else
		{
			temp.tv_sec = stop.tv_sec-start.tv_sec;
			temp.tv_nsec = stop.tv_nsec-start.tv_nsec;
		}
	} while( !temp.tv_sec && temp.tv_nsec < 250000000 );
	//timeout case, we need reset
//Timeout1:
	if (uiVerbosity >=  LOGGING_ERRORS)
		slogf( _SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_WARNING,
				"%s:i2c_master_wait_status: Timeout, reset, return 0 (CTRREG 0x%X STSREG 0x%X)",
				pszProgram, in16(dev->regbase + I2C_CTRREG_OFF), in16(dev->regbase + I2C_STSREG_OFF));
	i2c_reset(dev);
	return 0;
}
#endif

const char	szWrite[] = "write";
const char	szRead[] = "read";

uint32_t i2c_slave_wait_forever(i2c_dev_t *dev, int interrupt)
{
	int	status, ctrl;
	int			iid, interr = EOK;
 //   int statusold = 0;
    char *	pszMode;

    if ( iSlave == interrupt )
    {
    	iSlave = 0;
    	out16(dev->regbase + I2C_CTRREG_OFF, CTRREG_IEN );
    }
    switch( interrupt )
    {
    case AM335x_I2C_XRDY:
    	iid = dev->slave_write_iid;
    	pszMode = szWrite;
    	break;
    case AM335x_I2C_RRDY:
    	iid = dev->slave_read_iid;
    	pszMode = szRead;
    	break;
    default:
    	return I2C_STATUS_ERROR;
	}
//	printf( "%s Status = 0x%X\n", pszMode, in16(dev->regbase + I2C_STSREG_OFF) );
	/* Wait forever until we get an interrupt */
	if (uiVerbosity >= LOGGING_DEBUG_FINE)
		slogf(_SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_DEBUG2, "%s:i2c_slave_wait_forever: %s entry iid %d (CTRREG 0x%X STSREG 0x%X)",
				pszProgram, pszMode, iid, in16(dev->regbase + I2C_CTRREG_OFF), in16(dev->regbase + I2C_STSREG_OFF));
	/*Wait for an interrupt event */
Retry:
	while ( iSlave || !( ( status = in16(dev->regbase + I2C_STSREG_OFF)) & STSREG_IAAS) ||
			!( ( ctrl  = in16(dev->regbase + I2C_CTRREG_OFF) ) & CTRREG_IEN) || iMaster ||
			(ctrl & CTRREG_MSTA ) || !( status & interrupt ) )
	{
		struct timespec request;
		request.tv_sec  = 0;
		request.tv_nsec = 2000; // 2 microseconds (although it is always at least 3 milliseconds)
		clock_nanosleep(CLOCK_MONOTONIC , 0, &request, NULL );
/*
		usleep(2);
		interr = InterruptWait_r(0, NULL);
		if (uiVerbosity >=  LOGGING_DEBUG_FINE)
			slogf( _SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_DEBUG2,
					"%s:i2c_slave_wait_forever:InterruptWait_r: 0x%X %s", pszProgram, interr, strerror(interr));
*/
		ctrl = in16(dev->regbase + I2C_CTRREG_OFF );
		if ( iMaster || (ctrl & CTRREG_MSTA ) )
		{
			interr = EOK;
			if (uiVerbosity >=  LOGGING_DEBUG_FINE)
				slogf( _SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_DEBUG2,
						"%s:i2c_slave_wait_forever: %s Interrupt for master so retry", pszProgram, pszMode);
			continue;
		}
		if ( !( ctrl & CTRREG_IEN) )
			continue;
		if ( !iMaster && !iSlave && ( status & AM335x_I2C_ERROR) )
		{
			if (uiVerbosity >= LOGGING_ERRORS)
				slogf(_SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_WARNING, "%s:i2c_slave_wait_forever: error return 0x%X", pszProgram, status);
			i2c_reset( dev );
			continue;
		}
/*
		if ( statusold != status )
		{
			printf( "%s INTSET = 0x%X INTCLR = 0x%X Control = 0x%X Status = 0x%X\n", pszMode, in16(dev->regbase + AM335x_I2C_IE_SET), in16(dev->regbase + AM335x_I2C_IE_CLR), ctrl, statusold = status );
			if ( ( status & AM335x_I2C_BB ) && !( ctrl & ( AM335x_MST | AM335x_TRX ) ) )
			{
				ctrl = in16(dev->regbase + AM335x_I2C_DATA ) & 0x0FF;
				printf( "%s Just read 0x%X\n", pszMode, ctrl );
				if ( !( in16( dev->regbase + AM335x_I2C_SBLOCK ) & AM335x_OAE0 ) )
				{
					printf( "%s OAE0 bit not set so setting\n", pszMode );
//	       			out16(dev->regbase + AM335x_I2C_SBLOCK, 0 );			// Tell it to only use OA0
	       			out16(dev->regbase + AM335x_I2C_SBLOCK, AM335x_OAE0 );			// Tell it to only use OA0
	       		}
	       	}
		}
*/
	}
	/* Clear the interrupt */
	/* If an interrupt is pending */
//	InterruptUnmask(dev->intr, iid);
//ss		out16(dev->regbase + I2C_STSREG_OFF, in16(dev->regbase + I2C_STSREG_OFF) & AM335x_I2C_INT_MASK );
//		out16(dev->regbase + I2C_STSREG_OFF, in16(dev->regbase + I2C_STSREG_OFF) & AM335x_I2C_RESET ) ; //AM335x_I2C_INT_MASK );
	//	out16(dev->regbase + I2C_STSREG_OFF, in16(dev->regbase + I2C_STSREG_OFF) & ~(STSREG_IIF));
	if ( iSlave )
		goto Retry;
	iSlave = interrupt;
	if (uiVerbosity >= LOGGING_DEBUG_FINE)
		slogf(_SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_DEBUG2, "%s:i2c_slave_wait_forever: %s return 0 (CTRREG 0x%X STSREG 0x%X)",
				pszProgram, pszMode, in16(dev->regbase + I2C_CTRREG_OFF), in16(dev->regbase + I2C_STSREG_OFF));
	return 0;
}

static inline i2c_status_t i2c_wait_busy(i2c_dev_t *dev)
{
	//wait for 500us
	int timeout = 5000;
	if (uiVerbosity >= LOGGING_DEBUG_FINE)
		slogf(_SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_DEBUG2, "%s:i2c_wait_busy: entry (CTRREG 0x%X STSREG 0x%X)",
				pszProgram, in16(dev->regbase + I2C_CTRREG_OFF), in16(dev->regbase + I2C_STSREG_OFF));
	while(!(in16(dev->regbase + I2C_STSREG_OFF) & STSREG_IBB) && timeout--)
	{
//		if (uiVerbosity >= LOGGING_DEBUG_FINE)			// This is overkill
//			slogf(_SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_DEBUG2, "%s:i2c_wait_busy: STSREG 0x%X", pszProgram, reg);
		nanospin_ns(100);
	}
	if(timeout<=0)
	{
		if (uiVerbosity >= LOGGING_DEBUG_FINE)
			slogf(_SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_DEBUG2,
					"%s:i2c_wait_busy: timed out (CTRREG 0x%X STSREG 0x%X), reset and return I2C_STATUS_ERROR",
					pszProgram, in16(dev->regbase + I2C_CTRREG_OFF), in16(dev->regbase + I2C_STSREG_OFF));
		i2c_reset(dev);
		return (I2C_STATUS_ERROR);
	}
	if (uiVerbosity >= LOGGING_DEBUG_FINE)
		slogf(_SLOGC_SNELL_IQMIT_I2C_DRIVER, _SLOG_DEBUG2, "%s:i2c_wait_busy: (CTRREG 0x%X STSREG 0x%X) return 0",
				pszProgram, in16(dev->regbase + I2C_CTRREG_OFF), in16(dev->regbase + I2C_STSREG_OFF) );
	return (0);
}

__SRCVERSION( "$URL: http://svn/product/branches/6.5.0/trunk/hardware/i2c/i2c/wait.c $ $Rev: 408944 $" );
